#!/usr/bin/python3
# -*- coding: utf-8 -*-
import gc

import apt
import apt.progress.base as apb
import apt_pkg
import dbus
import locale
import os
from apt.cache import FetchFailedException
from apt.debfile import DebPackage
from kylindbus.enums import check_apt_lock, JSON_PKG_NAME_LIST
from kylindbus.log_manager import logger


class AppActions:
    def __init__(self):
        pass

    INS_AND_UP = "install_and_upgrade"
    INSTALL = "install"
    INSTALL_ONE = "insone"
    INSTALL_DEBFILE = "install_debfile"
    REMOVE_DEBFILE = "remove_debfile"
    UPGRADE = "upgrade"
    UPGRADE_ONE = "upone"
    UPDATE = "update"
    UPDATESHELL = "update_by_shell"
    CONFIGURESHELL = "configure_dpkg_by_shell"
    FIXCONFFILE = "fix_conffile_by_shell"


class FetchProcess(apb.AcquireProgress):
    taskPkgs = []

    def __init__(self, dbus_service, appname, action, pkgs):
        apb.AcquireProgress.__init__(self)
        self.percent = 0.0
        self.dbus_service = dbus_service
        self.appname = appname
        self.action = action
        self.pkgname = ''.join(pkgs)

    #
    # 函数:下载结束
    # item:下载的item
    #
    def done(self, item):
        kwarg = {"download_appname": self.appname,
                 "download_percent": str(self.percent),
                 "action": str(self.action),
                 "download_pkgname": str(self.pkgname),
                 }

        # 一个包下载成功，从任务列表中删除
        onePkgName = item.description.split()[3]
        if onePkgName in self.taskPkgs:
            self.taskPkgs.remove(onePkgName)

        self.dbus_service.kum_fetch_signal("down_done", kwarg)

    #
    # 函数:下载失败
    # item:下载的item
    #
    def fail(self, item):
        kwarg = {"download_appname": self.appname,
                 "download_percent": str(self.percent),
                 "action": str(self.action),
                 }

        self.dbus_service.kum_fetch_signal("down_fail", kwarg)

    #
    # 函数:接收数据
    # item:下载的item
    #
    def fetch(self, item):
        kwarg = {"download_appname": self.appname,
                 "download_percent": str(self.percent),
                 "action": str(self.action),
                 }

        self.dbus_service.kum_fetch_signal("down_fetch", kwarg)

    #
    # 函数:接触到ims
    # item:下载的item
    #
    def ims_hit(self, item):
        # 'ims_hit'
        pass

    #
    # 函数:媒介切换
    # media:媒介
    #
    def media_change(self, media, drive):
        # 'media_change'
        pass

    #
    # 函数:下载心跳状态
    #
    def pulse(self, owner):
        if self.action == AppActions.UPDATE:
            if self.total_items != 0:
                percent = float(self.current_items * 100.0 / self.total_items)
                if percent > self.percent:
                    self.percent = percent
        else:
            if self.total_bytes != 0:
                self.percent = float(self.current_bytes * 100.0 / self.total_bytes)

        kwarg = {"download_appname": self.appname,
                 "download_percent": str(self.percent),
                 "download_bytes": str(self.current_bytes),
                 "total_bytes": str(self.total_bytes),
                 "download_items": str(self.current_items),
                 "total_items": str(self.total_items),
                 "action": str(self.action),
                 "download_pkgname": str(self.pkgname),
                 }

        self.dbus_service.kum_fetch_signal("down_pulse", kwarg)

        # cancel the operation
        # if self.dbus_service.check_cancel_worker_item(self.appname) is True:
        # ("download_cancel")
        # self.dbus_service.kum_fetch_signal("down_cancel", kwarg)
        # return False

    #
    # 函数:开始下载
    #
    def start(self):
        # Reset all our values.
        self.current_bytes = 0.0
        self.current_cps = 0.0
        self.current_items = 0
        self.elapsed_time = 0
        self.fetched_bytes = 0.0
        self.last_bytes = 0.0
        self.total_bytes = 0.0
        self.total_items = 0
        self.percent = 0

        kwarg = {"download_appname": self.appname,
                 "download_percent": str(self.percent),
                 "download_items": str(self.total_items),
                 "action": str(self.action),
                 "download_pkgname": str(self.pkgname),
                 }

        self.dbus_service.kum_fetch_signal("down_start", kwarg)

    #
    # 函数:下载停止
    #
    def stop(self):
        kwarg = {"download_appname": self.appname,
                 "download_percent": str(100),
                 "action": str(self.action),
                 }

        # 有包下载失败但是未触发error，大多是因为依赖不满足导致，手动发送error信号
        # if self.total_bytes == 0.0 or self.current_bytes == 0.0:
        #     kwarg = {"apt_appname": self.appname,
        #              "apt_percent": "-99",
        #              "action": str(self.action),
        #              "status": "deps_error",
        #              }
        #     self.dbus_service.kum_apt_signal("down_error", kwarg)
        # else:
        self.dbus_service.kum_fetch_signal("down_stop", kwarg)


class AptProcess(apb.InstallProgress):
    """Apt progress"""

    def __init__(self, dbus_service, appname, action, count, pkg_name):
        apb.InstallProgress.__init__(self)
        self.dbus_service = dbus_service
        self.appname = appname
        self.percent = 0
        self.action = action
        self.install_num = count
        self.pkgname = ''.join(pkg_name)
        self.install_status = True

    #
    # 函数:冲突处理
    #
    def conffile(self, current, new):
        kwarg = {"apt_appname": str(current),
                 "apt_percent": str(6),
                 "error_message": str(str(current) + "与" + str(new) + "安装冲突"),
                 }
        self.install_status = False
        logger.error("安装状态：失败 - 包名：" + current + " - 错误码：(6) - 详情：" + str(str(current) + "与" + str(new) + "安装冲突"))
        self.dbus_service.kum_apt_signal("apt_error", kwarg)

    #
    # 函数:错误处理
    #
    def error(self, pkg, errormsg):
        kwarg = {"apt_appname": str(pkg),
                 "apt_percent": str(7),
                 "error_message": str(errormsg),
                 }
        if "/var/cache/apt/archives" in pkg:
            pkg = os.path.basename(pkg).split("_")[0]
        if pkg == self.appname:
            self.install_status = False
            logger.error("安装状态：失败 - 包名：" + pkg + " - 错误码：(7) - 详情：" + errormsg)
            self.dbus_service.error_message = errormsg

    #
    # 函数:开始处理
    #
    def start_update(self):
        kwarg = {"apt_appname": self.appname,
                 "apt_percent": str(self.percent),
                 "action": str(self.action),
                 }
        self.dbus_service.kum_apt_signal("apt_start", kwarg)

    #
    # 函数:处理结束
    #
    def finish_update(self):
        if self.install_status == True and self.percent != 0.0:
            kwarg = {"apt_appname": self.appname,
                     "apt_percent": str(100),
                     "action": str(self.action),
                     }
            logger.info("安装状态：成功 - 包名：" + self.appname)
            self.dbus_service.kum_apt_signal("apt_finish", kwarg)  # 安装完成信号
            self.dbus_service.insert_app_message("apt_finish", kwarg)  # 安装完成写入数据库
            self.dbus_service.counter -= 1
        elif self.install_status == False and self.percent == 0.0:  # percnet=0.0 兼容考虑静默升级依赖不存在时进度条会是0.0
            kwarg = {"apt_appname": self.appname,
                     "apt_percent": str(0),
                     "action": str(self.action),
                     }
            logger.info("安装状态：失败 - 包名：" + self.appname + "依赖解析失败")
            # self.dbus_service.kum_apt_signal("apt_error", kwarg)  # 安装完成信号
            # self.dbus_service.insert_app_message("apt_error", kwarg)  # 安装完成写入数据库
        elif self.install_status == True and self.percent == 0.0:  # 已被安装
            kwarg = {"apt_appname": self.appname,
                     "apt_percent": str(100),
                     "action": str(self.action),
                     }
            # logger.info("安装状态：成功 - 包名：" + self.appname + "已安装")
            # self.dbus_service.kum_apt_signal("apt_finish", kwarg)  # 安装完成信号
            # self.dbus_service.insert_app_message("apt_finish", kwarg)  # 安装完成写入数据库
        else:
            self.install_status = True

    #
    # 函数:处理过程
    #
    def status_change(self, pkg, percent, status):
        kwarg = {"apt_appname": self.appname,
                 "apt_percent": str(percent),
                 "action": str(self.action),
                 "status": str(status),
                 }
        self.dbus_service.kum_apt_signal("apt_pulse", kwarg)


class AptHandler:

    def __init__(self, dbus_service):
        self.dbus_service = dbus_service
        locale.setlocale(locale.LC_ALL, "zh_CN.UTF-8")
        self.cache = apt.Cache()
        # self.init_cache()
        # self.cache = apt.Cache()
        # self.cache.open()

    #
    # 函数:初始化cache
    #
    def init_cache(self):
        self.cache = apt.Cache()
        try:
            self.cache.open()
        except:
            logger.critical("初始化cache失败！！！", exc_info=True)

    #
    # 函数:删除cache
    #
    def delete_cache(self):
        if self.cache is not None:
            self.cache.close()
            del self.cache
            gc.collect()

    #
    # 函数:获取软件包
    # pkgName:软件包名
    #
    def get_pkg_by_name(self, pkgName):
        try:
            return self.cache[pkgName]
        except KeyError:
            raise WorkitemError(49, "由于源被更改或其它原因，导致 %s 无法在源中检测到" % pkgName)

    #
    # 函数：遍历列表判断是否依赖存在问题，以及标记包成安装和升级状态
    #
    def traverse_list(self, pkg_name, list):
        self.init_cache()
        print("--->list:", list)
        print("--->pkg_name:", pkg_name)
        for i in list:
            if i["appname"] == pkg_name and not i["depends_status"]:
                kwarg = {"apt_appname": pkg_name,
                         "apt_percent": str(50),
                         "error_message": "获取依赖包失败"}
                self.dbus_service.insert_app_message("depends_error", kwarg)
                return True
            elif i["appname"] == pkg_name and i["depends_status"]:
                if i["json"] == 2:  # 需要走解析json方式，标记所有的包
                    for app in i["install_list"]:
                        pkg = self.get_pkg_by_name(app)
                        pkg.mark_install()
                    for app in i["upgrade_list"]:
                        pkg = self.get_pkg_by_name(app)
                        pkg.mark_upgrade()
                return False
        return False

    #
    # 函数:批量自动安装或升级
    # pkg_names:软件包名
    #
    def install_and_upgrade(self, pkg_name, kwargs=None):
        # 删除本地缓存的deb安装包
        # os.popen("rm /var/cache/apt/archives/*.deb").readlines()
        # self.cache.open()
        logger.info("开始安装----- %s -----", pkg_name)

        # 单独处理依赖存在问题的软件包
        if self.traverse_list(pkg_name, self.dbus_service.optional_message_list):
            return
        # self.traverse_list(self.dbus_service.silent_message_list)
        self.dbus_service.efi_status = self.dbus_service.get_efi_status()
        self.dbus_service.set_efi_status(True)
        # for pkg_name in pkg_names:
        if pkg_name not in JSON_PKG_NAME_LIST:
            pkg = self.get_pkg_by_name(pkg_name)
            if not pkg.is_installed:
                pkg.mark_install()
            elif pkg.is_upgradable:
                pkg.mark_upgrade()
            else:  # 防止当前包为其他包的依赖包，先作为依赖已安装了，此时直接过滤-lj
                kwarg = {"apt_appname": pkg_name,
                         "apt_percent": str(100),
                         "action": "install",
                         }
                logger.info("安装状态：已安装 - 包名：" + pkg_name)
                self.dbus_service.kum_apt_signal("apt_finish", kwarg)  # 安装完成信号
                self.dbus_service.insert_app_message("apt_finish", kwarg)  # 安装完成写入数据库
        try:
            self.cache.commit(FetchProcess(self.dbus_service, pkg_name, AppActions.UPGRADE_ONE, pkg_name),
                              AptProcess(self.dbus_service, pkg_name, AppActions.UPGRADE_ONE, 1, pkg_name))
        except apt.cache.FetchFailedException as error:  # 静默升级下载包时失败
            raise WorkitemError(5, str(error))
        except apt.cache.LockFailedException as error:  # 尝试cache锁失败
            raise WorkitemError(4, str(error))
        except apt_pkg.Error as error:  # dpkg被锁
            check_apt_lock()
            raise WorkitemError(3, str(error))
        except SystemError as error:
            if self.dbus_service.error_message != "":
                raise WorkitemError(2, str(self.dbus_service.error_message))  # 包缺陷导致异常
            else:
                raise WorkitemError(2, str(error))
        except Exception as error:  # 其它异常，未知
            raise WorkitemError(1, str(error))
        logger.info("安装结束----- %s ------", pkg_name)
        self.dbus_service.set_efi_status(False)
        self.cache.close()


class WorkitemError(Exception):

    def __init__(self, errornum, details=""):
        self.errornum = errornum
        self.details = details


if __name__ == "__main__":
    ad = AptHandler(None)
